/*********************************************************************/
/*      Licensed Materials - Property of IBM                         */
/*      "Restricted Materials of IBM"                                */
/*      5748-T15                                                     */
/*      (C) Copyright IBM Corp. 2006                                 */
/*      All Rights Reserved                                          */
/*      US Government Users Restricted Rights - Use, duplication     */
/*      or dislosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                               */
/*********************************************************************/
/* This file is provided on an "as is" basis.  There are no          */
/* warranties, express or implied, including the implied warranties  */
/* of merchantability and fitness for a particular purpose.          */
/* IBM has no obligation to provide service, defect correction, or   */
/* any maintenance for the file.  IBM has no obligation to supply    */
/* any updates or enhancements for the file to you even if such are  */
/* or later become available.                                        */
/*********************************************************************/
/* Change History                                                    */
/* D14165  03/15/06   Update help information                        */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/* GENERAL INFORMATION:                                              */
/*                                                                   */
/* Segment Name:  qm5a00.c                                           */
/*                                                                   */
/* Abstract:      Member of the QM5A CSO which comprises the         */
/*                TPF GCC MEM5 driver.                               */
/*                                                                   */
/* Description:   This segment is the main entry point for the MEM5  */
/*                driver.  It will perform parsing and invoke the    */
/*                appropriate routine.                               */
/*                                                                   */
/*                This segment works with a table-driven design.     */
/*                A table entry exists for each macro/C-function,    */
/*                API or group of APIs that is supported by this     */
/*                driver.                                            */
/*                The table entry consists of the following:         */
/*                                                                   */
/*                      - name of the API                            */
/*                      - pointer to routine to assembler routine    */
/*                      - pointer to C++ function                    */
/*                                                                   */
/* ECB fields                                                        */
/* ebx000(1)  error case indicator                                   */
/* ebx001(1)  last iteration indicator                               */
/* ebx002(1)  -all api used                                          */
/* ebcm01(1)  verbose indicator                                      */
/*                                                                   */
/*********************************************************************/
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <tpf/sysapi.h>
#include <tpf/tpfapi.h>
#include <tpf/tpfio.h>
#include <tpf/tpfparse.h>
#include <tpf/i_decb.h>
#include <tpf/c_stck.h>
#include <qm5a70.h>

#define QM5A_GRAMMAR "{API-caa+++ [(NO)ERR] [(NO)Verbose] {[CNT-d++++] | [TIme-d+++]} [(NO)CHILD]}"

/*-------------------------------------------------------------------*/
/* function prototype                                                */
/*-------------------------------------------------------------------*/
boolean getapi(char *api, unsigned char *errcase, int *count, int *seconds, unsigned char *verbose, boolean *childECB);

/*-------------------------------------------------------------------*/
/*          MEM5 Driver functions                                    */
/*-------------------------------------------------------------------*/
void ECBHeap_func(void);
void ECBHeap2_func(void);
void systemheap_func(void);
void systemheap2_func(void);
void SWB_func(void);
void MixedHeap_func(void);
void LongRunningHeap_func(void);
void ECBMax_func(void);
void MallInfo_func(void);

void CreateQM5A(char *zmsg);

 /*-------------------------------------------------------------------*/
/* structure definition                                              */
/*-------------------------------------------------------------------*/

typedef struct api_tbl {
            char           api_name[9];
            void           (*api_asm)(void); /* ptr to assem func    */
            void           (*api_c)(void);   /* ptr to C func        */
            } API_TBL;

#define NUM_OF_APIS  8

const API_TBL  api_tbl[] =
{
/*    api   name           assembler rtn    c function               */
/*********************************************************************/
/*               MEM5 Driver APIS                                    */
/*********************************************************************/
   {  "EHEAP",             NULL,            ECBHeap_func },
   {  "SHEAP",             NULL,            systemheap_func },
   {  "SHEAP2",            NULL,            systemheap2_func },
   {  "SWB",               NULL,            SWB_func   },
   {  "MIXED",             NULL,            MixedHeap_func },
   {  "LONG",              NULL,            LongRunningHeap_func },
   {  "MAX",               NULL,            ECBMax_func },
   {  "MALL",              NULL,            MallInfo_func },
   { "" }
};

/*********************************************************************/
/*                                                                   */
/* QM5A( )                                                           */
/*                                                                   */
/* Entry point to the ZTEST MEM5  functional message.                */
/*                                                                   */
/*********************************************************************/
void QM5A(void)
{
   int                  i;
   int                  rc;
   char                 api[9] = { 0 };
   char                 help[255] = { 0 };
   char                 spaces[5] = { 0x40 };
   char                *buffer_ptr;
   boolean              found = FALSE;
   int                  count = 1;
   int                  seconds = 0;
   tpf_TOD_type         timeNow = {0};
   int                  startTime = 0;
   int                  currentTime = 0;
   boolean              childECB = FALSE;
   char                 zmsg[250] = "";
   struct wtopc_header  header;                                              //D15235

//   ecbptr()->ebrout = 0x010000;         //commented out for d15235 to cause
                                          //messages to route to the original termainal
   ecbptr()->ebx000 = FALSE;              /*default:  error cases = no    */
   ecbptr()->ebx001 = FALSE;              /*default:  last iteration = no */
   ecbptr()->ebx002 = FALSE;              /*default:  api-all = no        */
   ecbptr()->ebcm01 = TRUE;               /*default:  verbose = yes       */

   if ( getapi(api, &(ecbptr()->ebx000), &count, &seconds, &(ecbptr()->ebcm01), &childECB) == FALSE )
   {
/*********************************************************************************/
/*                           help display                                        */
      wtopc_insert_header(&header, NULL, 0, 'I', WTOPC_SYS_TIME);                              //D15235
      sprintf(help, "\n FORMAT:  ZTEST i MEM5 API-name \n"                                     //F14165
              "         (NO)ERR (NO)Verbose CNT-count TIme-seconds \n"                         //F14165
              " \n API-name is the only required parameter, specify ALL or a specific API. \n" //F14165
              " \n The following parameters are optional:");                                   //F14165
      rc = wtopc(help, WTOPC_EBROUT, WTOPC_PAGE, &header);                                     //D15235
      if ( rc == WTOPC_PAGE_STOP )
         exit(0);
      memset (help, 0x00, sizeof(help));
      sprintf(help, " NOERR is the default, specify ERR to exercise error conditions. \n"      //F14165
              " Verbose is the default, specify NOVerbose to display only error messages. \n"  //F14165
              " CNT-count or TIme-seconds may be specified to repeat the request. \n\n"        //F14165
              " Macros and functions supported by MEM5: \n");
      rc = wtopc(help, WTOPC_EBROUT, WTOPC_PAGE, WTOPC_NO_HEADER);
      if ( rc == WTOPC_PAGE_STOP )
         exit(0);
      memset (spaces,' ', sizeof(spaces));
      count = 0;
      memset (help, 0x00, sizeof(help));
      buffer_ptr = help;
      for ( i = 0; i < NUM_OF_APIS; i++ )
      {
         if ( buffer_ptr >= (help + sizeof(help) - sizeof(api) - sizeof(spaces) -count - 2 ))
         {
            rc = wtopc(help, WTOPC_EBROUT, WTOPC_PAGE, WTOPC_NO_HEADER);
            if ( rc == WTOPC_PAGE_STOP )
               exit(0);
            memset (help, 0x00, sizeof(help));
            count = 0;
            buffer_ptr = help;
         }
         buffer_ptr += sprintf (buffer_ptr, " %-6.6s%.5s",                                     //F14165
                                api_tbl[i].api_name,
                                spaces);
         ++count;
         if ( count == 4 )
         {
            memcpy(buffer_ptr, "\n", 1);
            ++buffer_ptr;
            count = 0;
         }
      }

      buffer_ptr += sprintf(buffer_ptr, " %-5s", "ALL");                                       //F14165
      rc = wtopc(help, WTOPC_EBROUT, WTOPC_PAGE, WTOPC_NO_HEADER);
      if ( rc == WTOPC_PAGE_STOP )
         exit(0);
      wtopc("End of MEM5 help display", WTOPC_EBROUT, WTOPC_PAGE_END, WTOPC_NO_HEADER);
      exit(0);
   }
/*********************************************************************************/

   tpf_STCK (&timeNow); /* get the value of TOD clock */
   if ( seconds )
      count = 2;

   /* use left most 4 bytes of TOD clock as an approximation of      */
   /* a "seconds" counter                                            */
   memcpy (&startTime, &timeNow, sizeof(startTime));

   if ( memcmp(api, "ALL", 3) == 0 )    /* execute cases for all api */
   {
       ecbptr()->ebx002 = TRUE;
       if (ecbptr()->ebx000)
          prtmsg("executing error cases for all API \n");
       else
          prtmsg("executing normal cases for all API \n");

       while (count > 0)
       {
           /*  if this is the only or last iteration, set the indicator  */
           /*    and run all the api test cases                          */
           if ( count == 1 )
           {
              ecbptr()->ebx001 = TRUE;

              for ( i = 0; i < NUM_OF_APIS; i++ )
              {
                 defrc();
                 if ( api_tbl[i].api_asm )
                 {
                    (*api_tbl[i].api_asm)();
                 }
                 defrc();
                 if ( api_tbl[i].api_c )
                 {
                    (*api_tbl[i].api_c)();
                 }
              }
           }
           else
           {
              /* If not, create a new child ecb to run the test cases */
              sprintf(zmsg, "ZTEST MEM5 API-%s %s %s CHILD", api,
                            ecbptr()->ebx000 ? "ERR":"NOERR",
                            ecbptr()->ebcm01 ? "VERBOSE":"NOVERBOSE");
              CreateQM5A(zmsg);
           }

           /* if the user specified an amount of time that the       */
           /* driver should run for, calculate if that time has      */
           /* expired yet.  If it has, decrement the count variable  */
           /* so we will be forced out of the while loop; else,      */
           /* continue in the loop.                                  */
           if (seconds > 0)
           {
               tpf_STCK (&timeNow);
               memcpy (&currentTime, &timeNow, sizeof(currentTime));
               if (currentTime - startTime >= seconds)
               {
                  /* time is up, force the next iteration to be the last */
                  seconds = 0;
                  count = 1;
               }
           }
           /* if the number of seconds to run was not specified,     */
           /* just decrement the count variable so we know whether   */
           /* or not to continue in the loop                         */
           else
               count--;
       }

       found = TRUE;
    }
    else
    {                           /* execute specific API test cases   */
       for ( i = 0; i < NUM_OF_APIS; i++ )
       {
          if ( memcmp(api, api_tbl[i].api_name, strlen(api)) == 0)
          {
              found = TRUE;

              while (count > 0)
              {
                  /*  if this is the only or last iteration, set the indicator  */
                  /*    and run the specific api test cases                     */
                  if ( count == 1 )
                  {
                     ecbptr()->ebx001 = TRUE;
                     if ( api_tbl[i].api_asm )
                     {
                        (*api_tbl[i].api_asm)();
                     }
                     if ( api_tbl[i].api_c )
                     {
                        (*api_tbl[i].api_c)();
                     }
                  }
                  else
                  {
                     /* If not, create a new child ecb to run the test cases */
                     sprintf(zmsg, "ZTEST MEM5 API-%s %s %s CHILD", api,
                               ecbptr()->ebx000 ? "ERR":"NOERR",
                               ecbptr()->ebcm01 ? "VERBOSE":"NOVERBOSE");
                     CreateQM5A(zmsg);
                  }

                  /* if the user specified an amount of time that*/
                  /* the driver should run for, calculate if that*/
                  /* time has expired yet.  If it has, decrement */
                  /* the count variable so we will be forced out */
                  /* of the while loop; else, just release all   */
                  /* DECBs and continue in the loop.             */
                  if (seconds > 0)
                  {
                      tpf_STCK (&timeNow);
                      memcpy (&currentTime, &timeNow,
                              sizeof(currentTime));
                      if (currentTime - startTime >= seconds)
                      {
                         /* time is up, force the next iteration to be the last */
                         seconds = 0;
                         count = 1;
                      }
                  }
                  /* if the number of seconds to run was not     */
                  /* specified, just decrement the count variable*/
                  /* so we know whether or not to continue in the*/
                  /* loop                                        */
                  else
                      count--;
              }
              break;

          } /* end if the matching API is found */
       }  /* end for loop that searches specific API */

       if ( ! found )
       {
          prtmsg(api);
          prtmsg(":  is not in the API table, no test cased executed \n");
       }
    } /* end executing specific API test cases */

   /* If this is the parent ECB, or VERBOSE mode is ON, issue message when it completes */
   if (!(childECB) || (ecbptr()->ebcm01))
      wtopc("MEM5 driver exited", WTOPC_EBROUT, WTOPC_CHAIN_END, WTOPC_NO_HEADER);

   exit(0);
}


/*********************************************************************/
/*                                                                   */
/* getapi( )                                                         */
/*                                                                   */
/* Routine to parse the input functional message and extract the     */
/* name of the api   which is to be executed and the error case      */
/* number if supplied.                                               */
/*                                                                   */
/*********************************************************************/
boolean getapi(char *api, unsigned char *errcase, int *count, int *seconds, unsigned char *verbose,
               boolean *childECB)
{
   int    i;
   struct IPRSE_output  parsed_data, *pparm;
   struct mi0mi        *blk_ptr;
   char                *input_ptr, *eom_ptr;
   int                  parser_return;

   blk_ptr = (struct mi0mi *)ecbptr()->ce1cr0;
   input_ptr = blk_ptr->mi0acc+strlen("Zxxxx");
   eom_ptr = (char *)&blk_ptr->mi0ln0 + blk_ptr->mi0cct - 1;
   *eom_ptr = '\0';

   parser_return = IPRSE_parse(input_ptr,
                              QM5A_GRAMMAR,
                              &parsed_data,
                              IPRSE_ALLOC+IPRSE_PRINT,
                              "QM5A");

   /******************************************************************/
   /* return -1 indicates HELP or ?                                  */
   /* return  0 indicates message did not meet grammar requirements  */
   /******************************************************************/

   if ( parser_return == -1 )
   {
      return(FALSE);
   }
   if ( parser_return == 0 )
   {
      return(FALSE);
   }

   /******************************************************************/
   /* Else, determine which parameters were input and set the        */
   /* appropriate values to return to the caller so the corresponding*/
   /* APIs can be tested.                                            */
   /******************************************************************/
   for ( i = 1, pparm = &parsed_data;
         i <= parser_return;
         i++ , pparm = pparm->IPRSE_next )
   {
      if ( memcmp(pparm->IPRSE_parameter, "API", 3) == 0 )
      {
         memcpy(api, pparm->IPRSE_value, strlen(pparm->IPRSE_value));
      }
      else if ( strcmp(pparm->IPRSE_parameter, "(NO)ERR") == 0 )
           {
               *errcase = *(pparm->IPRSE_value) == 'Y' ? TRUE : FALSE;
           }
      else if ( memcmp(pparm->IPRSE_parameter, "CNT", 3) == 0 )
           {
               *count = atoi(pparm->IPRSE_value);
           }
      else if ( memcmp(pparm->IPRSE_parameter, "TIme", 4) == 0 )
           {
               *seconds = atoi(pparm->IPRSE_value);
           }
      else if ( strcmp(pparm->IPRSE_parameter, "(NO)Verbose") == 0 )
           {
               *verbose = *(pparm->IPRSE_value) == 'Y' ? TRUE : FALSE;
           }
      else if ( strcmp(pparm->IPRSE_parameter, "(NO)CHILD") == 0 )
           {
               *childECB = *(pparm->IPRSE_value) == 'Y' ? TRUE : FALSE;
           }
   }

   return(TRUE);
}

/*********************************************************************/
/*                                                                   */
/* CreateQM5A                                                        */
/*                                                                   */
/* Routine to create a child ECB to run the MEM5 test cases.         */
/* tpf_cresc is used to create a new child ECB, and wait for its     */
/* completion before it returns back to the caller.                  */
/*                                                                   */
/*********************************************************************/
void CreateQM5A(char *zmsg)                      // create ecb_count ECBs
{
  struct mi0mi* block_ptr = NULL;

  struct tpf_cresc_input tci;
  struct tpf_ev0bk_list_data *ebptr;     /* Pointer to Event block    */
                                         /* with list item data       */

  tpf_ev0bk_list_item *liptr;            /* Pointer to list item data */

  block_ptr = (struct mi0mi *) malloc(sizeof(struct mi0mi));

  block_ptr->mi0cct = strlen(zmsg) + 4;          // set message length
  strcpy(block_ptr->mi0acc, zmsg);               // put functional message on the core block

  tci.program = "CVZZ";                 /* Pgm for child to enter     */
  tci.istream = TPF_CRESC_IS_MAIN;      /* I/S for child              */
  tci.timeout = 300;                    /* Timeout value for parent   */
                                      /* Parent will wait forever   */
  tci.wait = TPF_CRESC_WAIT_YES;        /* Create all children        */
  tci.data_length = sizeof(struct mi0mi);
  tci.data = block_ptr;

  ebptr = tpf_cresc(&tci);

  liptr = TPF_EVENT_LIST_ITEM(ebptr,0);


  if ((liptr->evnbklif & TPF_EVNBK_POST) != 0)  /* Item posted?       */
     if ((liptr->evnbklif & TPF_EVNBK_ECDE) != 0) /* Yes, posted    */
                                              /* with error?        */
     {
          prterr("MEM5: Child exited with error");  /* Yes, issue msg.         */
     }
     else
     {
          prtmsg("MEM5: Child ECB exits"); /* Yes, successful       */
     }
  else
      prterr("MEM5: CHILD ECB is not posted because the timeout value was exceeded.");

  if (block_ptr)
     free(block_ptr);
}
