/*********************************************************************/
/*      Licensed Materials - Property of IBM                         */
/*      "Restricted Materials of IBM"                                */
/*      5748-T15                                                     */
/*      (C) Copyright IBM Corp. 2006                                 */
/*      All Rights Reserved                                          */
/*      US Government Users Restricted Rights - Use, duplication     */
/*      or dislosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                               */
/*********************************************************************/
/* This file is provided on an "as is" basis.  There are no          */
/* warranties, express or implied, including the implied warranties  */
/* of merchantability and fitness for a particular purpose.          */
/* IBM has no obligation to provide service, defect correction, or   */
/* any maintenance for the file.  IBM has no obligation to supply    */
/* any updates or enhancements for the file to you even if such are  */
/* or later become available.                                        */
/*********************************************************************/
/* Change History                                                    */
/*                                                                   */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/* GENERAL INFORMATION:                                              */
/*                                                                   */
/* Segment Name:  qm5a70.c                                           */
/*                                                                   */
/* Abstract:      Member of the QM5A CSO which comprises the         */
/*                TPF GCC MEM5 driver.                               */
/*                                                                   */
/* Description:   This segment contains MEM5 driver tests:           */
/*                                                                   */
/*                - ECB Heap                                         */
/*                - SWBs                                             */
/*                - mixed memory usage:                              */
/*                     -- ECB Heap                                   */
/*                     -- System Heap                                */
/*                     -- SWBs                                       */
/*                     -- C++ new and delete operators               */
/*                     -- traditional storage (getcc and relcc)      */
/*                - long running, looping tests:                     */
/*                     -- ECB Heap                                   */
/*                     -- System Heap                                */
/*                - ECB Max                                          */
/*                - Mallinfo                                         */
/*                                                                   */
/*********************************************************************/
#include <string.h>
#include <stdlib.h>
#include <tpf/tpfapi.h>
#include <tpf/c_ck1ke.h>
#include <stdio.h>
#include <unistd.h>
#include <tpf/c_eb0eb.h>
#include <tpf/c_proc.h>
#include <tpf/sysapi.h>
#include <limits.h>        // for INT_MAX
#include <tpf/i_ecb3.h>    // for tpf_ecbp3()
#include <errno.h>
#include <qm5a70.h>
#include <qm5a80.h>
#include <tpf/malloc.h>    // for mallinfo()
#include <sys/time.h>      // for gettimeofday()

//---------------------------------------------------------------------
// function prototypes
//---------------------------------------------------------------------

void systemheap_func(void);
void systemheap2_func(void);
int  malloc_func(enum t_heap);
int  calloc_func(enum t_heap);
int  realloc_func(enum t_heap);
int  realloc2_func(void);
void *getMalloc(size_t, enum t_heap);
void *getCalloc(size_t, size_t, enum t_heap);
void *getRealloc(void *, size_t, enum t_heap);
void FreeStorage(void *);
void free_func();
int  writeToStorage(void *, int, int);
int  readFromStorage(void *, int);
int  doErrorPaths(enum t_heap);
int  checkNulls(char *, int, int);
void SWB_func();
void cgswbc_func();
void *get_SWB(char *);
void release_SWB(void *);
void getcc_func();
void relcc_func();
void attac_levels(unsigned char *);
void detac_levels(unsigned char *);
int  LongRunningHeap_func(void);

boolean  expectedError = FALSE;
char ECBHeapAbortMsg[] =
         "aborting further allocation tests due to errors";

/*********************************************************************/
/* Function Name:   ECBHeap()                                        */
/*********************************************************************/
void ECBHeap_func(void)
{
    int rc = 0;

    msgPrefix = "ECBHeap";

    if (DoErrors) {
         expectedError = TRUE;
         if ((rc = doErrorPaths(TPF_HEAP_31))) {
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         if ((rc = doErrorPaths(TPF_HEAP_64))) {
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

    } else {
         if ((rc = malloc_func(TPF_HEAP_31))) {
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         if ((rc = malloc_func(TPF_HEAP_64))) {
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         if ((rc = calloc_func(TPF_HEAP_31))) {
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         if ((rc = calloc_func(TPF_HEAP_64))) {
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         if ((rc = realloc_func(TPF_HEAP_31))){
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         if ((rc = realloc_func(TPF_HEAP_64))){
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         if ((rc = realloc2_func())){           // mix 31 & 64
             PrintError(ECBHeapAbortMsg);
             return;
         }
         defrc();

         free_func();
    }
    PrintMsg("completed ECBHeap_func");
}

/*********************************************************************/
/* Function Name:   SWB()                                            */
/*********************************************************************/
void SWB_func(void)
{
    msgPrefix = "SWB";
    if (DoErrors)
    {
        PrintMsg("No error test cases defined for SWB_func");
        return;
    }
    cgswbc_func();
    PrintMsg("completed SWB_func");
}


/*********************************************************************/
/* Function Name:   MixedHeap()                                      */
/*********************************************************************/
void MixedHeap_func(void)
{
    unsigned char detach = 0;
    msgPrefix = "MIXED Heap";

    if (DoErrors)
    {
        PrintMsg("No error test cases defined for MixedHeap_func");
        return;
    }

    detac_levels(&detach);
    defrc();

    getcc_func();
    defrc();

    cgswbc_func();
    defrc();

    systemheap2_func();
    defrc();

    systemheap_func();
    defrc();

    relcc_func();
    defrc();

    attac_levels(&detach);
    defrc();

    msgPrefix = "MIXED Heap";
    PrintMsg("completed MixedHeap_func");
}

/*********************************************************************/
/* Function Name:   LongRunningHeap_func()                           */
/*********************************************************************/
int LongRunningHeap_func(void)
{
    int rand_pick, rand_seed, num, size, i, rc=0;
    static int addr3_frames=0, addr5_frames=0, addr7_frames=0, addr11_frames=0;
    char token3[9], token5[9], token7[9], token11[9];
    struct timeval now;
    static void *addr0_ptr=NULL,                 // malloc
                *addr1_ptr=NULL,                 // calloc
                *addr2_ptr=NULL,                 // system heap
                *addr3_ptr=NULL,                 // malloc
                *addr4_ptr=NULL,                 // realloc
                *addr5_ptr=NULL,                 // calloc
                *addr6_ptr=NULL,                 // system heap
                *addr7_ptr=NULL,                 // calloc
                *addr8_ptr=NULL,                 // realloc
                *addr9_ptr=NULL,                 // calloc
                *addr10_ptr=NULL,                // realloc
                *addr11_ptr=NULL;                // system heap

    msgPrefix = "Long Running Heap";

    if (DoErrors)
    {
        PrintMsg("No error test cases defined for LongRunningHeap_func");
        return(0);
    }

    if (ecbptr()->ebx001 == FALSE) {
        PrintMsg("Malloc different storage sizes for testing");
        // test case : malloc size of 4567
        if (!addr0_ptr) {                 // already released ?
             size = 4567;
             if (!(addr0_ptr = getMalloc(size, TPF_HEAP_31)))
                 return(1);
             if ((rc = writeToStorage(addr0_ptr, size, 1)))
                 return(1);
             if ((rc = readFromStorage (addr0_ptr, size)))
                 return (1);
        }

        // test case : calloc size of 4000, number elements 1
        if (!addr1_ptr) {                 // already released ?
            size = 4000;
            if (!(addr1_ptr = getCalloc(1, size, TPF_HEAP_64)))
                return(1);
             if ((rc = writeToStorage(addr1_ptr, size, 1)))
                 return(1);
             if ((rc = readFromStorage (addr1_ptr, size)))
                 return (1);
        }

        // test case : malloc size of 25000
        if (!addr2_ptr) {                 // already released ?
             size = 25000;
             if (!(addr2_ptr = getMalloc(size, TPF_HEAP_64)))
                 return(1);
             if ((rc = writeToStorage(addr2_ptr, size, 1)))
                 return(1);
             if ((rc = readFromStorage (addr2_ptr, size)))
                 return (1);
        }

        // test case :  system heap
        addr3_frames = 2;
        memcpy (token3, "xyzdefgh", 8);
        if (!addr3_ptr) {                 // already released ?
            if (!(addr3_ptr = GetSystemHeap(addr3_frames, token3, perm[0].owner, perm[0].flags)))
                 return(1);
            rc = WriteToSystemHeap(addr3_ptr, addr3_frames, perm[0].flags);
            if (rc) {
                ReturnSystemHeap(addr3_ptr,
                                 addr3_frames,
                                 token3,
                                 perm[0].owner,
                                 perm[0].flags);
                return(1);
            }
        }

        // test case : calloc size of 4095, number elements 5
        if (!addr4_ptr) {                 // already released ?
            size = 4095;
            num = 5;
            if (!(addr4_ptr = getCalloc(num, size, TPF_HEAP_64)))
                return(1);
            if ((rc = writeToStorage(addr4_ptr, size * num, 1)))
                return(1);
        }

        // test case :  system heap
        addr5_frames = 1;
        memcpy (token5, "87654321", 8);
        if (!addr5_ptr) {                 // already released ?
            if (!(addr5_ptr = GetSystemHeap(addr5_frames, token5, perm[12].owner, perm[12].flags)))
                 return(1);
            rc = WriteToSystemHeap(addr5_ptr, addr5_frames, perm[12].flags);
            if (rc) {
                ReturnSystemHeap(addr5_ptr,
                                 addr5_frames,
                                 token5,
                                 perm[12].owner,
                                 perm[12].flags);
                addr5_ptr = NULL;
                return(1);
            }
        }

        // test case :  calloc num 1000,  array elements sizeof(char*)
        if (!addr6_ptr) {                 // already released ?
            num = 1000;
            size = sizeof(char *);
            if (!(addr6_ptr = getCalloc(num, size, TPF_HEAP_31)))
                return(1);
            if ((rc = writeToStorage(addr6_ptr, size * num, 1)))
                return(1);
        }

        // test case :  get system heap
        addr7_frames = 7;
        memcpy (token7, "tokenid1", 8);
        if (!addr7_ptr) {                 // already released ?
            if (!(addr7_ptr = GetSystemHeap(addr7_frames, token7, perm[5].owner, perm[5].flags)))
                 return(1);
            rc = WriteToSystemHeap(addr7_ptr, addr7_frames, perm[5].flags);
            if (rc) {
                ReturnSystemHeap(addr7_ptr,
                                 addr7_frames,
                                 token7,
                                 perm[5].owner,
                                 perm[5].flags);
                return(1);
            }

        }

        // test case :  malloc size of 1
        if (!addr8_ptr) {                 // already released ?
            size = 1;
            if (!(addr8_ptr = getMalloc(size, TPF_HEAP_31)))
                return(1);
            if ((rc = writeToStorage(addr8_ptr, size, 1)))
                return(1);
            if ((rc = readFromStorage (addr8_ptr, size)))
                return (1);
        }

        // test case : malloc 150,000
        if (!addr9_ptr) {                 // already released ?
            size = 150000;
            if (!(addr9_ptr = getMalloc(size, TPF_HEAP_64)))
                return(1);
            if ((rc = writeToStorage(addr9_ptr, size, 1)))
                return(1);
            if ((rc = readFromStorage (addr9_ptr, size)))
                return (1);
        }

        // test case : calloc large size
        if (!addr10_ptr) {                // already released ?
            num = 1;
            size = 100000;
            if (!(addr10_ptr = getCalloc(num, size, TPF_HEAP_64)))
                return(1);
            if ((rc = writeToStorage(addr10_ptr, size * num, 1)))
                return(1);
        }

        // test case :  get system heap
        addr11_frames = 3;
        memcpy (token11, "another1", 8);
        if (!addr11_ptr) {                // already released ?
            if (!(addr11_ptr = GetSystemHeap(addr11_frames, token11, perm[8].owner, perm[8].flags)))
                return(1);
            rc = WriteToSystemHeap(addr11_ptr, addr11_frames, perm[8].flags);
            if (rc) {
                ReturnSystemHeap(addr11_ptr,
                                 addr11_frames,
                                 token11,
                                 perm[8].owner,
                                 perm[8].flags);
                return(1);
            }
        }

        MixedHeap_func();

        msgPrefix = "Long Running Heap";

        // get a random number
        for (i = 0; i < 6; i++) {
             now.tv_sec  = 0;
             now.tv_usec = 0;
             gettimeofday(&now, NULL);
             rand_seed = now.tv_usec;
             srand(rand_seed);
             rand_pick = rand();
             rand_pick = rand_pick % 12;

             PrintMsg("Select and run a random test case");

             switch (rand_pick) {
                case 0:
                    if (addr0_ptr) {
                        PrintMsg("Free a malloc storage");
                        FreeStorage(addr0_ptr);
                        addr0_ptr = NULL;
                    }
                    break;
                case 1:
                    if (addr1_ptr) {
                        PrintMsg("Free a malloc storage");
                        FreeStorage(addr1_ptr);
                        addr1_ptr = NULL;
                    }
                    break;
                case 2:
                    if (addr2_ptr) {
                        PrintMsg("Free a malloc storage");
                        FreeStorage(addr2_ptr);
                        addr2_ptr = NULL;
                    }
                    break;
                case 3:
                    if (addr3_ptr) {
                        ReturnSystemHeap(addr3_ptr,
                                         addr3_frames,
                                         token3,
                                         perm[0].owner,
                                         perm[0].flags);
                        addr3_ptr = NULL;
                     }
                     break;
                case 4:
                    if (addr4_ptr) {
                        FreeStorage(addr4_ptr);
                        addr4_ptr = NULL;
                    }
                    break;
                case 5:
                    if (addr5_ptr) {
                        ReturnSystemHeap(addr5_ptr,
                                         addr5_frames,
                                         token5,
                                         perm[12].owner,
                                         perm[12].flags);
                        addr5_ptr = NULL;
                    }
                    break;
                case 6:
                    if (addr6_ptr) {
                        FreeStorage(addr6_ptr);
                        addr6_ptr = NULL;
                    }
                    break;
                case 7:
                    if (addr7_ptr) {
                        ReturnSystemHeap(addr7_ptr,
                                         addr7_frames,
                                         token7,
                                         perm[5].owner,
                                         perm[5].flags);
                        addr7_ptr = NULL;
                    }
                    break;
                case 8:
                    if (addr8_ptr) {
                        FreeStorage(addr8_ptr);
                        addr8_ptr = NULL;
                    }
                    break;
                case 9:
                    if (addr9_ptr) {
                        FreeStorage(addr9_ptr);
                        addr9_ptr = NULL;
                    }
                    break;
                case 10:
                    if (addr10_ptr) {
                        FreeStorage(addr10_ptr);
                        addr10_ptr = NULL;
                    }
                    break;
                case 11:
                    if (addr11_ptr) {
                        ReturnSystemHeap(addr11_ptr,
                                         addr11_frames,
                                         token11,
                                         perm[8].owner,
                                         perm[8].flags);
                        addr11_ptr = NULL;
                    }
                    break;
                default:
                    PrintMsg("unknown case");
             }  // switch
        }  // for
    } else {
          PrintMsg("Test case completed. Releasing all storage.");
          //  release everything
          if (addr0_ptr)
          {
              FreeStorage(addr0_ptr);
              addr0_ptr = NULL;
          }
          if (addr1_ptr)
          {
              FreeStorage(addr1_ptr);
              addr1_ptr = NULL;
          }
          if (addr2_ptr)
          {
              FreeStorage(addr2_ptr);
              addr2_ptr = NULL;
          }
          if (addr3_ptr)
          {
              addr3_frames = 2;
              memcpy (token3, "xyzdefgh", 8);
              ReturnSystemHeap(addr3_ptr,
                               addr3_frames,
                               token3,
                               perm[0].owner,
                               perm[0].flags);
              addr3_ptr = NULL;
          }
          if (addr4_ptr)
          {
              FreeStorage(addr4_ptr);
              addr4_ptr = NULL;
          }
          if (addr5_ptr)
          {
              addr5_frames = 1;
              memcpy (token5, "87654321", 8);
              ReturnSystemHeap(addr5_ptr,
                               addr5_frames,
                               token5,
                               perm[12].owner,
                               perm[12].flags);
              addr5_ptr = NULL;
          }
          if (addr6_ptr)
          {
              FreeStorage(addr6_ptr);
              addr6_ptr = NULL;
          }
          if (addr7_ptr)
          {
              addr7_frames = 7;
              memcpy (token7, "tokenid1", 8);
              ReturnSystemHeap(addr7_ptr,
                               addr7_frames,
                               token7,
                               perm[5].owner,
                               perm[5].flags);
              addr7_ptr = NULL;
          }
          if (addr8_ptr)
          {
              FreeStorage(addr8_ptr);
              addr8_ptr = NULL;
          }
          if (addr9_ptr)
          {
              FreeStorage(addr9_ptr);
              addr9_ptr = NULL;
          }
          if (addr10_ptr)
          {
              FreeStorage(addr10_ptr);
              addr10_ptr = NULL;
          }
          if (addr11_ptr)
          {
              addr11_frames = 3;
              memcpy (token11, "another1", 8);
              ReturnSystemHeap(addr11_ptr,
                               addr11_frames,
                               token11,
                               perm[8].owner,
                               perm[8].flags);
              addr11_ptr = NULL;
          }
     }

     PrintMsg("completed LongRunningHeap_func");

     return (0);
}

/*********************************************************************/
/* Function Name:      ECBMax_func()                                 */
/*                     f2708  test DCR 64b-067                       */
/*                     Test to allow application programs to raise   */
/*                     their ecb maximum allowed values              */
/*********************************************************************/
void ECBMax_func(void)
{
       int max31;      // maximum size of 31-bit heap
       int max64;      // maximum size of 64-bit heap

       static int MAX64 = 0;      // maximum size of 64-bit heap
       int size;       // size of desired storage
       int rc;
       struct mallinfo info;
       int bufferSize;

       void *malloc1_ptr = NULL;
       void *malloc2_ptr = NULL;

       msgPrefix = "ECB Max";
       PrintMsg("entering ECBMax_func");

       struct iproc *iproc_ptr = iproc_process;        // access ecb page 2 for ce2proc
                                                       // gets the value of 31 and 64 bit heap

       // store the starting ecb maximums
       max31 = iproc_ptr->iproc_maxheap_31;    // store max size of 31-bit heap (MMES)
       max64 = iproc_ptr->iproc_maxheap_64;    // store max size of 64-bit heap (XMMES)

       if (!MAX64)
           MAX64 = max64;     /* save original value  */
       /*********************************************/
       /* Create parmlists for tpf_ebmaxc function  */
       /*********************************************/

       // parmlist to increase 31 and 64 bit heap to the maximum the system will allow
       struct ebmaxc_parmlist myparm1[] =  {
                       { MAX31HEAP , MAXVALUE },
                       { MAX64HEAP , MAXVALUE },
                       { 0 , 0 }
       };

       // parmlist for error case 2 and normal case 2
       // incriment 31 and 64 bit heap maximums
       struct ebmaxc_parmlist myparm2[] =  {
                       { MAX31HEAP , 1 },
                       { MAX64HEAP , 2 },
                       { 0 , 0 }
       };

       // parmlist for normal case 3
       // increase maximum values by ugreater than the system will allow
       struct ebmaxc_parmlist myparm3[] =  {
                       { MAX31HEAP , (MAXVALUE + 1) },
                       { MAX64HEAP , (MAXVALUE + 1) },
                       { 0 , 0 }
       };

       // run error cases
       if( (DoErrors) )
       {
               expectedError = TRUE;   // doesnt print unsuccessful malloc messages

               PrintMsg("Running Error Cases For ECBMax_func");

               /*****************************************/
               /* error case 1 : use system default max */
               /*****************************************/

               // attempt to allocate greater than maximum
               size = (max31 + 1) * (ONE_MEG);
               if ((malloc1_ptr = getMalloc(size, TPF_HEAP_31)))
               {
                       PrintError("error case 1 : allocated space greater than 31-bit heap max");
                       FreeStorage(malloc1_ptr);
               }

               size = (max64 + 1) * (ONE_MEG);
               if ((malloc2_ptr = getMalloc(size, TPF_HEAP_64)))
               {
                       PrintError("error case 1 : allocated space greater than 64-bit heap max");
                       FreeStorage(malloc2_ptr);
               }

               PrintMsg("Creating child ECB to run error ECB Private Area case, should cause a CTL-I0006DB");
               CreateECB_int(MAX_ECB, 2);

               /***************************************************/
               /* error case 2 : increase max with tpf_ebmaxc     */
               /*                max value should be increased    */
               /*                only for MAX31HEAP and MAX64HEAP */
               /***************************************************/

               tpf_ebmaxc(myparm2);

               // set new max
               max31 = iproc_ptr->iproc_maxheap_31;
               max64 = iproc_ptr->iproc_maxheap_64;

               // attempt to allocate greater than maximum
               size = (max31 + 1) * (ONE_MEG);
               if ((malloc1_ptr = getMalloc(size, TPF_HEAP_31)))
               {
                       PrintError("error case 2 : allocated space greater than 31-bit heap max");
                       FreeStorage(malloc1_ptr);
               }

               size = (max64 + 1) * (ONE_MEG);
               if ((malloc2_ptr = getMalloc(size, TPF_HEAP_64)))
               {
                       PrintError("error case 2 : allocated space greater than 64-bit heap max");
                       FreeStorage(malloc2_ptr);
               }

               /***************************************************/
               /* error case 3 : increase max by less than or     */
               /*                equal to 0. This should fail and */
               /*                a dump                           */
               /***************************************************/

               PrintMsg("Creating child ECB to run error case, should cause a dump");
               CreateECB_int(MAX_ECB, 3);

               /****************************************************/
               /* error case 4 : increase max to the system max    */
               /****************************************************/

               tpf_ebmaxc(myparm1);

               // set new max
               max31 = iproc_ptr->iproc_maxheap_31;
               max64 = iproc_ptr->iproc_maxheap_64;

               // attempt to allocate greater than maximum
               size = (max31 + 1) * (ONE_MEG);
               if ((malloc1_ptr = getMalloc(size, TPF_HEAP_31)))
               {
                       PrintError("error case 4 : allocated space greater than 31-bit heap max");
                       FreeStorage(malloc1_ptr);
               }

               size = (max64 + 1) * (ONE_MEG);
               if ((malloc2_ptr = getMalloc(size, TPF_HEAP_64)))
               {
                       PrintError("error case 4 : allocated space greater than 64-bit heap max");
                       FreeStorage(malloc2_ptr);
               }

               PrintMsg("Completed Error Cases for ECBMax_func");
       }

       // run normal cases
       else
       {
               PrintMsg("Running Normal Cases for ECBMax_func");

               /***********************************/
               /* case 1 : use system default max */
               /***********************************/

               info = mallinfo();
               bufferSize = info.fordblks; //max

               if (bufferSize > 0) { // max31 is max # 1MB
                   if (!(malloc1_ptr = getMalloc(bufferSize, TPF_HEAP_31)))
                       PrintError("case 1 : unable to allocate space for 31-bit heap");
               // allocate maximum amount of space allowed
               }

               size = (MAX64 - 2) * ONE_MEG;    // subract 2MB to allow for used heap
               if (size > 0)
               {
                  if (!(malloc2_ptr = getMalloc(size, TPF_HEAP_64)))
                       PrintError("case 1 : unable to allocate space for 64-bit heap");
               }

               // free all storage
               if (malloc1_ptr) {	
                   FreeStorage(malloc1_ptr);
                   malloc1_ptr = NULL;	
               }
               if (malloc2_ptr)	{
                   FreeStorage(malloc2_ptr);
                   malloc2_ptr = NULL;	
               }
               PrintMsg("Creating child ECB to run normal ECB Private Area case");
               CreateECB_int(MAX_ECB, 1);
               sleep(1);

               /*********************************************/
               /* case 2 : increase max with tpf_ebmaxc     */
               /*          value should be increased        */
               /*          only for MAX31HEAP and MAX64HEAP */
               /*          MAXEPRIV done in QM5B.C          */
               /*********************************************/

               tpf_ebmaxc(myparm2);

               // compare the new max value to the start max value
   	        rc = iproc_ptr->iproc_maxheap_31;
               if(max31 < rc)
               {
                    info = mallinfo();
                    bufferSize = info.fordblks;  //  max

                    if (bufferSize > 0) {
                       if (!(malloc1_ptr = getMalloc(bufferSize, TPF_HEAP_31)))
                           PrintError("case 2 : unable to allocate space for 31-bit heap");
                    }
               }

               rc = iproc_ptr->iproc_maxheap_64;
               size = 2 * ONE_MEG;    // get the additional storage we allowed to get
               if (!(malloc2_ptr = getMalloc(size, TPF_HEAP_64)))
                   PrintError("case 2 : unable to allocate space for 64-bit heap");

               // free all storage
               if (malloc1_ptr) {	
                   FreeStorage(malloc1_ptr);
                   malloc1_ptr = NULL;	
               }
               if (malloc2_ptr)	{
                   FreeStorage(malloc2_ptr);
                   malloc2_ptr = NULL;	
               }

               // increase values to maximums
               tpf_ebmaxc(myparm1);

               // set new max
               max31 = iproc_ptr->iproc_maxheap_31;
               max64 = iproc_ptr->iproc_maxheap_64;


               /***********************************************/
               /* case 3 : increase max greater than MAXVALUE */
               /*          value should remain the same       */
               /*          Only for MAX31HEAP and MAX64HEAP   */
               /*          MAXEPRIV done in QM5B.C            */
               /***********************************************/

               tpf_ebmaxc(myparm3);

               // compare the new max value to the start max value
               // should be the same
               if(max31 != (rc = iproc_ptr->iproc_maxheap_31))
               {
                       sprintf(msgBuffer, "case 3 : 31-bit max changed from %d (MB) to %d (MB)", max31, rc);
                       PrintError(msgBuffer);
               }
               if(max64 != (rc = iproc_ptr->iproc_maxheap_64))
               {
                       sprintf(msgBuffer, "case 3 : 64-bit max changed from %d (MB) to %d (MB)", max64, rc);
                       PrintError(msgBuffer);
               }

               PrintMsg("Completed Normal Cases for ECBMax_func");
       }

       PrintMsg("completed ECBMax_func");
       return;
}

/*********************************************************************/
/* Function Name:   MallInfo_func()                                  */
/*                  f2265  test DCR 64b-059                          */
/*                  create new API to test mallinfo() function       */
/*                  only works with the 31 bit heap                  */
/*********************************************************************/
void MallInfo_func(void)
{
       int rc;
       int max31;
       int fsm_temp;   // temp storage for fsmblks
       int bufferSize;
       long *ptrArray;
       void *malloc1_ptr = NULL;
       int i = 0;

       struct mallinfo info;   // access data from mallinfo()
       struct iproc *iproc_ptr = iproc_process;        // access ecb page 2 for ce2proc

       msgPrefix = "MallInfo";
       PrintMsg("Entering MallInfo_func");

       // parameter list for tpf_ebmaxc function
       // do this so that EMPS and MMES are the same value
       struct ebmaxc_parmlist myparm[] =  {
                       { MAX31HEAP , MAXVALUE },
                       { 0, 0 }
       };

       max31 = iproc_ptr->iproc_maxheap_31;

       // There are no error cases for mallinfo. Return back to caller if ERR is specified.
       if (DoErrors)
       {
          PrintMsg("No error cases defined for MallInfo");
          return;
       }

       info = mallinfo();      // get all info from the mallinfo function

       /*********************************/
       /* CASE 1 : ordblks and keepcost */
       /*********************************/

       PrintMsg("case 1: testing ordblks and keepcost");

       // get storage for an array of pointers for malloc we allocated.
       bufferSize = 10240;
       ptrArray = getCalloc(bufferSize, 1, TPF_HEAP_31);

       info = mallinfo();
       // while the number of 1mb frames in used is not at maximum
       while (info.keepcost < info.ordblks)
       {
          // allocate the largest space we can
          bufferSize = info.fordblks;
          // If there is no more free blocks to malloc, or malloc fails,
          // there is an error somewhere.
          if ((bufferSize <= 0) || ((ptrArray[i] = (long) getMalloc(bufferSize, TPF_HEAP_31)) == 0))
             break;
          i++;
          info = mallinfo();
       }

       // check if the test case completes successfully or fails
       if( (info.keepcost) != (info.ordblks) )
       {
          sprintf(msgBuffer, "case 1 failed: ordblks = %d, keepcost = %d", info.ordblks, info.keepcost);
          PrintError(msgBuffer);
       }
       else
       {
          sprintf(msgBuffer, "case 1 successful: ordblks = %d, keepcost = %d", info.ordblks, info.keepcost);
          PrintMsg(msgBuffer);
       }

       // Done with the test case. Release all storage we malloc.
       i = 0;
       while (ptrArray[i])
       {
           FreeStorage((void *)ptrArray[i]);
           i++;
       }
       FreeStorage(ptrArray);


       /********************************/
       /* CASE 2 : usmblks and fsmblks */
       /********************************/

       info = mallinfo();

       PrintMsg("case 2: testing usmblks and fsmblks");

       rc = (info.fsmblks);
       fsm_temp = (info.usmblks);

       if (rc > 0) // Allocate if there are some free blocks
       {
          if(!(malloc1_ptr = getMalloc(rc, TPF_HEAP_31)))
               PrintError("case 2: unable to allocate space from 31 bit heap");
          else
          {
               rc = (rc + fsm_temp);

               info = mallinfo();

               // total size, in use, from main memory plus an 8 byte corruption buffer
               if( (info.usmblks) != (rc+8) )
               {
                       sprintf(msgBuffer, "case 2: original usmblks = %d, original fsmblks = %d, new usmblks = %d", fsm_temp, (rc - fsm_temp), info.usmblks);
                       PrintError(msgBuffer);
               }

               FreeStorage(malloc1_ptr);
               malloc1_ptr = NULL;
          }
       }

       /*************************************************************/
       /* CASE 3 : fordblks and uordblks                            */
       /* fordblks = maximum malloc size guaranteed by the system.  */
       /*************************************************************/

       info = mallinfo();

       PrintMsg("case 3: testing fordblks and uordblks");
       // Try to malloc what's guaranteed
       rc = info.fordblks;
       if (rc > 0)
       {
          if(!(malloc1_ptr = getMalloc(rc, TPF_HEAP_31)))
             // failed
             PrintError("case 3: unable to allocate space from 31 bit heap");
          else
          {
             // success
             info = mallinfo();

             // to make sure uordblks is incremented properly
             if ((info.uordblks) < rc)
             {
                 sprintf(msgBuffer, "case 3: uordblks = %d, fordblks = %d", info.uordblks, info.fordblks);
                 PrintError(msgBuffer);
             }

             // done.. free the storage
             FreeStorage(malloc1_ptr);
             malloc1_ptr = NULL;
          }
       }

       /******************************/
       /* CASE 4 : arena and smblks */
       /******************************/

       // raise the value of MMES so that it is equal to EMPS
       tpf_ebmaxc(myparm);

       max31 = iproc_ptr->iproc_maxheap_31;
       max31 = ( (ONE_MEG)*(max31) );

       info = mallinfo();      // get the updated mallinfo

       PrintMsg("case 4: testing arena and smbblks");
       rc = ( (max31) + ( (FOUR_KB)*(info.smblks) ) );
       if( (info.arena) != rc )
       {
               sprintf(msgBuffer, "case 4: arena = %d, (EMPS + 4KB * smblks) = %d", info.arena, rc);
               PrintError(msgBuffer);
       }

       /*****************************/
       /* CASE 5 : hblks and hblkhd */
       /*****************************/

       info = mallinfo();

       PrintMsg("case 5: testing hblks and hblkhd");

       /* hblks and hblkhd are fixed numbers right now */
       /* these values should change if the control table becomes larger */
       rc = ((info.hblks) * (FOUR_KB));
       if( ((info.hblks) != 5) && ((info.hblkhd) != rc) )
       {
               sprintf(msgBuffer, "case 5: hblks = %d, hblkhd = %d", info.hblks, info.hblkhd);
               PrintError(msgBuffer);
       }

       PrintMsg("completed MallInfo_func");
       return;
}


/*********************************************************************/
/* Function Name:   malloc_func()                                    */
/*********************************************************************/
int malloc_func(enum t_heap heap)
{
    int  size, rc;
    void *malloc1_ptr, *malloc2_ptr, *malloc3_ptr, *malloc4_ptr;
    long totalSpaceGotten = 0;
    long *ptrArray;
    char *tagName    = "HEAPTAG123",
         *tagName1   = "heaptag123",
         *dupTagName = "HEAPTAG1_dup_8Bytes";

    struct ctka         *kyptA;             // Pointer to Keypoint A
    unsigned short      havsz1, havsz2,
                        havsz3, havsz4;     // Size of buffs on ECB heap AVLn
    unsigned short      peh, mmes, xmmes;
    int                 i = 0;
    struct mallinfo info;   // access data from mallinfo()
    errno = 0;

    // Get Keypoint A. If there is an error retrieving
    // Keypoint A then output an error message and exit the ECB.

    // go to BSS to get keypoint A values
    cebic_save();
    cebic_goto_bss();

    if (tpf_getkc('A', (void *) &kyptA, GETKC_NONE) != 0)  {
       PrintError("Error retrieving Keypoint A ");
       return(1);
    }

    // store value from retrieved from keypoint A
    havsz1 = kyptA->ck1havsz1;
    havsz2 = kyptA->ck1havsz2;
    havsz3 = kyptA->ck1havsz3;
    havsz4 = kyptA->ck1havsz4;

    mmes   = kyptA->ck1mmes;
    xmmes  = kyptA->ck1xmmes;
    peh    = kyptA->ck1mcpeh(kyptA->ck1mcinus);

    // free the malloc space obtained by tpf_getkc()
    free(kyptA);

    // get back to the DBI where we were.
    cebic_restore();

    switch (heap) {
      case TPF_HEAP_31:
          PrintMsg("entering malloc_func for 31 bit ");
          break;
      case TPF_HEAP_64:
          PrintMsg("entering malloc_func for 64 bit ");

          /*************************/
          /* Test case for 64b-049 */
          /*************************/

          memset (msgBuffer, 0, sizeof(msgBuffer));
          sprintf (msgBuffer, "Got sizes of buffs on ECB heap AVLn:\n HAVSZ1-%d, "
                              "HAVSZ2-%d, HAVSZ3-%d, HAVSZ4-%d",
                              havsz1, havsz2, havsz3, havsz4);
          PrintMsg(msgBuffer);

          // test case for 64b-049 :  malloc size of havsz1 - 1
          size = havsz1 - 1;
          if (!(malloc1_ptr = getMalloc(size, heap)))
             return(1);
          FreeStorage(malloc1_ptr);

          // test case for 64b-049 :  malloc size of havsz2 - 1
          size = havsz2 - 1;
          if (!(malloc1_ptr = getMalloc(size, heap)))
             return(1);
          FreeStorage(malloc1_ptr);

          // test case for 64b-049 :  malloc size of havsz3 - 1
          size = havsz3 - 1;
          if (!(malloc1_ptr = getMalloc(size, heap)))
             return(1);
          FreeStorage(malloc1_ptr);

          // test case for 64b-049 :  malloc size of havsz4 - 1
          size = havsz4 - 1;
          if (!(malloc1_ptr = getMalloc(size, heap)))
             return(1);
          FreeStorage(malloc1_ptr);

          // test case for 64b-049 :  malloc size of havsz4 + 1
          size = havsz4 + 1;
          if (!(malloc1_ptr = getMalloc(size, heap)))
             return(1);
          FreeStorage(malloc1_ptr);

          /*************************/
          /* Test case for 64b-051 */
          /*************************/

          // APIs: tpf_eheap_tag() and tpf_eheap_locate()
          malloc4_ptr = getMalloc(size, heap);

	       // If we were able to get the storage, then tag the heap
      	 if (malloc4_ptr)
          {
             PrintMsg("Tests tpf_eheap_tag( ) ");
	          rc = tpf_eheap_tag (malloc4_ptr, tagName);
   		    if (rc != TPF_EHEAP_OK)  {
	      	    PrintError("Error tagging heap, return from here.");
                return(1);
             }

             // Error case 1
             PrintMsg("Tests error case: The first 8 bytes of the specified tag are not unique");
             malloc1_ptr = getMalloc(size, heap);
             if (malloc1_ptr)  {
    	          rc = tpf_eheap_tag (malloc1_ptr, dupTagName);
      		    if (rc != TPF_EHEAP_DUPETAG)  {
	         	    PrintError("Didn't get return code of TPF_EHEAP_DUPETAG, return from here.");
                   return(1);
                }
                FreeStorage(malloc1_ptr);

                // Error case 2
                PrintMsg("Tests error case: The ECB heap buffer address specified is not valid");
    	          rc = tpf_eheap_tag (malloc1_ptr + 1, tagName1);
      		    if (rc != TPF_EHEAP_ADDR_NOT_VALID)  {
	         	    PrintError("Didn't get return code of TPF_EHEAP_ADDR_NOT_VALID, return from here.");
                   return(1);
                }

                // Error case 3
                PrintMsg("Tests error case: The ECB heap buffer address specified already has a tag associated with it");
    	          rc = tpf_eheap_tag (malloc4_ptr, tagName1);
      		    if (rc != TPF_EHEAP_TAG_EXISTS)  {
	         	    PrintError("Didn't get return code of TPF_EHEAP_TAG_EXISTS, return from here.");
                   return(1);
                }

             }
       	    else
                PrintError("malloc1_ptr: Could not allocate heap ");

             PrintMsg("Testing tpf_eheap_locate(tagName) ");
             malloc4_ptr = tpf_eheap_locate (tagName);
	
             // Did we find the heap?  If so, release it
             if (malloc4_ptr != NULL)
                FreeStorage(malloc4_ptr);
             else
	             PrintError("Could not locate tag ");
          }
   	    else
	          PrintError("malloc4: Could not allocate heap ");

          /*********************************************/
          /* Error case: Try tpf_eheap_locate( ) again */
          /*********************************************/
          // Error case 4
          malloc4_ptr = tpf_eheap_locate (tagName);
	
          // Did we find the buffer?  If so, release it
          if (malloc4_ptr)  {
             PrintError("tpf_eheap_locate( ) didn't fail with released tag, return from here.");
             return(1);
          }
          break;
    }

    // test case :  malloc size of 1
    size = 1;
    if (!(malloc1_ptr = getMalloc(size, heap)))
        return(1);
    FreeStorage(malloc1_ptr);

    // test case :   malloc size of 4000
    size = SIZE_4K;
    if (!(malloc2_ptr = getMalloc(size, heap)))
        return(1);
    FreeStorage(malloc2_ptr);

    // test case :   malloc size of 64000, free
    size = SIZE_64K;
    if (!(malloc3_ptr = getMalloc(size, heap)))
        return(1);
    FreeStorage(malloc3_ptr);

    // test case :   malloc size of 256000, write to area
    size = SIZE_256K;
    if (!(malloc3_ptr = getMalloc(size, heap)))
        return(1);
    if ((rc = writeToStorage (malloc3_ptr, size, 1)))
        return(1);
    FreeStorage(malloc3_ptr);

    // test case :   malloc > pre-allocated size  100000
    // keypoint A, ctka, new value EMPSINIT (default is 16 - 4K)
    // 64kb initially carved out
    size = 100000;
    if (!(malloc3_ptr = getMalloc(size, heap)))
        return(1);
    FreeStorage(malloc3_ptr);

    // test case :   malloc max amt of ECB heap
    // EMPS is max size ecb heap in meg
    // MMHS/XMMHS were first named but then changed to MMES/XMMES
    // MMES  is max # 1M frames backing 31 bit
    // XMMES is max # 1M frames backing 64 bit

    switch (heap) {
      case TPF_HEAP_31:

           // Get some storage for the the array of addresses obtained by malloc
           size = 10240;
           malloc3_ptr = getCalloc(size, 1, heap);
           ptrArray = (long *) malloc3_ptr;

           // Get the keepcost value from mallinfo, to determine how many 1mb frames
           // have been used for the 31bit heap.
           // If the maximum (MMES) has not been reached yet, determine what's the
           // largest piece of memory we can malloc and allocate that.
           i = 0;
           info = mallinfo();      // get the updated mallinfo

           // Get how much 31-bit heap frames are used now.
           totalSpaceGotten = info.uordblks;

           // keep getting storage until we hit the maximum
           while (totalSpaceGotten/ONE_MEG < mmes)
           {
              size = info.fordblks;
              // If there is no more free blocks to malloc, or malloc fails,
              // there is an error somewhere.
              if ((size <= 0) || ((ptrArray[i] = (long) getMalloc(size, heap)) == 0))
                 break;
              i++;
              // keep a running total
              totalSpaceGotten += size;
              info = mallinfo();
           }

           // We cannot get the maximum 1mb frames allowed for 31bit heap for this ecb.
           // this test case fails.
           if (info.keepcost != mmes)
           {
              sprintf (msgBuffer,
                 "Failed to malloc max amount of 31-bit heap : KEEPCOST = %d and MMES = %d\n",
                  info.keepcost, mmes);
              PrintError(msgBuffer);
              return (1);
           }
           else
           {
              // We have already acquired the maximum 1mb frames for the 31bit ecb heap.
              // The following calloc should fail.
              expectedError = TRUE;
              if (malloc(ONE_MEG) != NULL)
              {
                 sprintf(msgBuffer,
                    "Able to malloc 1MB of storage when KEEPCOST = MMES = %dMB\n",
                    info.keepcost);
                 PrintError(msgBuffer);
                 return (1);
              }
              expectedError = FALSE;
           }

           // Done with the test case. Release all storage we malloc.
           i = 0;
           ptrArray = (long *) malloc3_ptr;
           while (ptrArray[i])
           {
              FreeStorage((void *)ptrArray[i]);
              i++;
           }
           FreeStorage(malloc3_ptr);
           break;

      case TPF_HEAP_64:
           // Try to test the max allowed 64bit heap only if we
           // can get more than 2 1-MB frames.
           if (xmmes > 2)
           {
              size = (xmmes - 2) * ONE_MEG;  // max64
              if (!(malloc3_ptr = getMalloc(size, heap)))
                 return(1);
              if ((rc = writeToStorage (malloc3_ptr, size, 1)))
                 return (1);
              if ((rc = readFromStorage (malloc3_ptr, size)))
                 return (1);
              FreeStorage(malloc3_ptr);
           }
           break;
    }

    return(0);

}

/*********************************************************************/
/* Function Name:   calloc_func()                                    */
/*********************************************************************/
int calloc_func(enum t_heap heap)
{

    int size;
    int num;
    int rc;
    void *calloc1_ptr, *calloc2_ptr, *calloc3_ptr;
    long *ptrArray;
    int i = 0;
    long   totalSpaceGotten = 0;
    struct mallinfo info;   // access data from mallinfo()

    struct ctka * kyptA = NULL;

    // go to BSS to get keypoint A values
    cebic_save();
    cebic_goto_bss();

    if (tpf_getkc('A', (void *) &kyptA, GETKC_NONE) != 0)  {
       PrintError("Error retrieving Keypoint A ");
       return(1);
    }

    // store value from retrieved from keypoint A
    unsigned short mmes   = kyptA->ck1mmes;
    unsigned short xmmes  = kyptA->ck1xmmes;

    // free the malloc space obtained by tpf_getkc()
    free(kyptA);

    // get back to the DBI where we were.
    cebic_restore();

    switch (heap) {
      case TPF_HEAP_31:
          PrintMsg("entering calloc_func for 31 bit ");
          break;
      case TPF_HEAP_64:
          PrintMsg("entering calloc_func for 64 bit ");
          break;
    }

    // test case : calloc size of 1, number elements 1
    size = 1;
    num  = 1;
    if (!(calloc1_ptr = getCalloc(num, size, heap)))
        return(1);
    if ((rc = checkNulls ((char *)calloc1_ptr, num, size)))
        return(1);
    FreeStorage(calloc1_ptr);

    // test case : calloc size of 5, number elements 5
    size = 5;
    num  = 5;
    if (!(calloc2_ptr = getCalloc(num, size, heap)))
        return(1);
    if ((rc = checkNulls ((char *)calloc2_ptr, num, size)))
        return(1);
    FreeStorage(calloc2_ptr);

    // test case : calloc size 1024, number elements 1
    num  = 1;
    size = SIZE_1K;
    if (!(calloc2_ptr = getCalloc(num, size, heap)))
        return(1);
    if ((rc = checkNulls ((char *)calloc2_ptr, num, size)))
        return(1);
    FreeStorage(calloc2_ptr);

    // test case : calloc size of 4000, number elements 1
    num  = 1;
    size = SIZE_4K;
    if (!(calloc2_ptr = getCalloc(num, size, heap)))
        return(1);
    if ((rc = checkNulls ((char *)calloc2_ptr, num, size)))
        return(1);
    FreeStorage(calloc2_ptr);

    // test case :  calloc size of 64000, number elements 1
    size = SIZE_64K;
    if (!(calloc3_ptr = getCalloc(num, size, heap)))
        return(1);
    FreeStorage(calloc3_ptr);

    // test case :  calloc size of 256000, num elements 1 and free
    size = SIZE_256K;
    if (!(calloc3_ptr = getCalloc(num, size, heap)))
        return(1);
    FreeStorage(calloc3_ptr);

    // test case :  calloc num 1000,  array elements sizeof(char*)
    num = 1000;
    size = sizeof(char *);
    if (!(calloc3_ptr = getCalloc(num, size, heap)))
        return(1);
    FreeStorage(calloc3_ptr);

    // test case :   calloc > pre-allocated size   100000
    // keypoint A, ctka, new value EMPSINIT (default is 16 - 4K)
    // 64kb initially carved out
    size = 100000;
    num = 1;
    if (!(calloc3_ptr = getCalloc(num, size, heap)))
        return(1);
    FreeStorage(calloc3_ptr);


    // test case :   calloc max amt of ECB heap
    // EMPS  is max size ecb heap in meg
    // MMHS/XMMHS were first named but then changed to MMES/XMMES
    // MMES  is max # 1M frames backing 31 bit
    // XMMES is max # 1M frames backing 64 bit
    switch (heap) {
      case TPF_HEAP_31:
           num = 1;
           // Get some storage for the the array of addresses obtained by malloc
           size = 10240;
           calloc3_ptr = getCalloc(size, num, heap);
           ptrArray = (long *) calloc3_ptr;

           // Get the keepcost value from mallinfo, to determine how many 1mb frames
           // have been used for the 31bit heap.
           // If the maximum (MMES) has not been reached yet, determine what's the
           // largest piece of memory we can malloc and allocate that.
           i = 0;
           info = mallinfo();      // get the updated mallinfo

           totalSpaceGotten = info.uordblks;
           while (totalSpaceGotten/ONE_MEG < mmes)
           {
              size = info.fordblks;
              // If there is no more free blocks to malloc, or Malloc fails,
              // there is an error somewhere.
              if ((size <= 0) || ((ptrArray[i] = (long) getCalloc(size, num, heap)) == 0))
                 break;
              i++;
              totalSpaceGotten += size;
              info = mallinfo();
           }

           // We cannot get the maximum 1mb frames allowed for 31bit heap for this ecb.
           // this test case fails.
           if (info.keepcost < mmes)
           {
              sprintf (msgBuffer,
                 "Failed to calloc max amount of 31-bit heap : KEEPCOST = %d and MMES = %d\n",
                  info.keepcost, mmes);
              PrintError(msgBuffer);
              return (1);
           }
           else
           {
              // We have already acquired the maximum 1mb frames for the 31bit ecb heap.
              // The following calloc should fail.
              expectedError = TRUE;
              if (getCalloc(ONE_MEG, num, heap) != NULL)
              {
                 sprintf(msgBuffer,
                    "Able to calloc 1MB of storage when KEEPCOST = %d, MMES = %dMB\n",
                    info.keepcost, mmes);
                 PrintError(msgBuffer);
                 return (1);
              }
              expectedError = FALSE;
           }

           // Done with the test case. Release all storage we malloc.
           i = 0;
           ptrArray = (long *) calloc3_ptr;
           while (ptrArray[i])
           {
              FreeStorage((void *) ptrArray[i]);
              i++;
           }
           FreeStorage(calloc3_ptr);
           break;

      case TPF_HEAP_64:
           if (xmmes > 2)
           {
              size = (xmmes - 2) * ONE_MEG;
              num = 1;
              if (!(calloc3_ptr = getCalloc(num, size, heap)))
                 return(1);
              if ((rc = checkNulls ((char *)calloc3_ptr, num, size)))
                 return(1);
              if ((rc = writeToStorage (calloc3_ptr, size * num, 1)))
                 return (1);
              if ((rc = readFromStorage (calloc3_ptr, size)))
                 return (1);
              FreeStorage(calloc3_ptr);
           }
           break;
    }
    return(0);
}

/*********************************************************************/
/* Function Name:   realloc_func()                                   */
/*********************************************************************/
int realloc_func(enum t_heap heap)
{
    int size;
    int num;
    int rc;
    void *malloc_ptr = NULL;
    void *calloc_ptr = NULL;
    void *realloc_ptr = NULL;
    long *ptrArray    = NULL;
    int  i = 0;
    struct mallinfo info;   // access data from mallinfo()

    struct ctka * kyptA = NULL;

    // go to BSS to get keypoint A values
    cebic_save();
    cebic_goto_bss();

    if (tpf_getkc('A', (void *) &kyptA, GETKC_NONE) != 0)  {
       PrintError("Error retrieving Keypoint A ");
       return(1);
    }

    // store value from retrieved from keypoint A
    unsigned short mmes   = kyptA->ck1mmes;
    unsigned short xmmes  = kyptA->ck1xmmes;

    // free the malloc space obtained by tpf_getkc()
    free(kyptA);

    // get back to the DBI where we were.
    cebic_restore();

    switch (heap) {
      case TPF_HEAP_31:
          PrintMsg("entering realloc_func for 31 bit ");
          break;
      case TPF_HEAP_64:
          PrintMsg("entering realloc_func for 64 bit ");
          break;
    }

    // test case : malloc, realloc with size of 1, free
    size = 150;
    if (!(malloc_ptr = getMalloc(size, heap)))
        return(1);
    size = 1;
    if (!(realloc_ptr = getRealloc(malloc_ptr, size, heap)))
        return(1);
    FreeStorage(realloc_ptr);


    // test case : malloc, realloc with larger size (1K), free
    size = 10;
    if (!(malloc_ptr = getMalloc(size, heap)))
        return(1);
    size = SIZE_1K;
    if (!(realloc_ptr = getRealloc(malloc_ptr, size, heap)))
        return(1);
    FreeStorage(realloc_ptr);

    // test case : malloc, realloc with smaller size
    size = 1000;
    if (!(malloc_ptr = getMalloc(size, heap)))
        return(1);
    size = 100;
    if (!(realloc_ptr = getRealloc(malloc_ptr, size, heap)))
        return(1);
    FreeStorage(realloc_ptr);


    // test case : calloc, realloc with same size (4K), free
    num = 2;
    size = SIZE_4K;
    if (!(malloc_ptr = getCalloc(num, size, heap)))
        return(1);
    size = SIZE_4K;
    if (!(realloc_ptr = getRealloc(malloc_ptr, size, heap)))
        return(1);
    FreeStorage(realloc_ptr);

    // test case : realloc with NULL pointer
    malloc_ptr = NULL;
    if (!(realloc_ptr = getRealloc(malloc_ptr, 5000, heap)))
        return(1);
    FreeStorage(realloc_ptr);


    // test case : calloc, write, realloc larger (64K), read, free
    size = 1000;
    num  = 1;
    if (!(calloc_ptr = getCalloc(num, size, heap)))
        return(1);
    if ((rc = writeToStorage (calloc_ptr, size * num, 1)))
        return (1);
    size = SIZE_64K;
    if (!(realloc_ptr = getRealloc(calloc_ptr, size, heap)))
        return(1);
    if ((rc = readFromStorage (realloc_ptr, 1000)))
        return (1);
    FreeStorage(realloc_ptr);


    // test case : calloc, realloc larger (256K), write, free
    size = 5000;
    num  = 1;
    if (!(calloc_ptr = getCalloc(num, size, heap)))
        return(1);
    size = SIZE_256K;
    if (!(realloc_ptr = getRealloc(calloc_ptr, size, heap)))
        return(1);
    if ((rc = writeToStorage(realloc_ptr, size, size)))
        return(1);
    FreeStorage(realloc_ptr);


    // test case :   realloc > pre-allocated size
    // keypoint A, ctka, new value EMPSINIT (default is 16 - 4K)
    // 64kb initially carved out
    num  = 1;
    size = 100000;
    if (!(calloc_ptr = getCalloc(num, size, heap)))
        return(1);
    size = 200000;
    if (!(realloc_ptr = getRealloc(calloc_ptr, size, heap)))
        return(1);
    FreeStorage(realloc_ptr);

    // test case :   malloc max amt of ECB heap
    // EMPS  is max size ecb heap in meg
    // MMHS/XMMHS were first named but then changed to MMES/XMMES
    // MMES  is max # 1M frames backing 31 bit
    // XMMES is max # 1M frames backing 64 bit
    switch (heap) {
      case TPF_HEAP_31:
           size = 10240;
           malloc_ptr = getMalloc(size, heap);
           ptrArray = (long *) malloc_ptr;

           // Get the keepcost value from mallinfo, to determine how many 1mb frames
           // have been used for the 31bit heap.
           // If the maximum (MMES) has not been reached yet, determine what's the
           // largest piece of memory we can malloc and allocate that.
           i = 0;
           info = mallinfo();      // get the updated mallinfo
           while (info.keepcost <= (mmes/2))
           {
              size = info.fordblks;
              // If there is no more free blocks to malloc, or Malloc fails,
              // there is an error somewhere.
              if ((size <= 0) || ((ptrArray[i] = (long) getMalloc(size, heap)) == 0))
                 break;
              i++;
              info = mallinfo();
           }

           // We cannot get the maximum 1mb frames allowed for 31bit heap for this ecb.
           // this test case fails.
           if (info.keepcost <= (mmes/2))
           {
              sprintf (msgBuffer,
                 "Failed to realloc max amount of 31-bit heap : KEEPCOST = %d and MMES = %d\n",
                  info.keepcost, mmes);
              PrintError(msgBuffer);
              return (1);
           }

           size = (info.fordblks > 10240) ? info.fordblks:10240;
           if (!(realloc_ptr = getRealloc(malloc_ptr, size, heap)))
           {
              sprintf(msgBuffer, "Realloc failed : KEEPCOST = %d, size = %d, MMES = %d\n",
                            info.keepcost, size, mmes);
              PrintError(msgBuffer);
              return(1);
           }

           // Done with the test case. Release all storage we malloc.
           i = 0;
           ptrArray = (long *) realloc_ptr;
           while (ptrArray[i])
           {
              FreeStorage((void *)ptrArray[i]);
              i++;
           }

           FreeStorage(realloc_ptr);
           break;

      case TPF_HEAP_64:
           if (xmmes > 2)
           {
              size = (xmmes - 2) * ONE_MEG / 2;
              if (!(malloc_ptr = getMalloc(size, heap)))
                 return(1);
              size += 1;
              if (!(realloc_ptr = getRealloc(malloc_ptr, size, heap)))
                 return(1);
              if ((rc = writeToStorage (realloc_ptr, size, size)))
                 return (1);
              if ((rc = readFromStorage (realloc_ptr, size)))
                 return (1);
              FreeStorage(realloc_ptr);
           }
           break;
    }
    return(0);
}
/*********************************************************************/
/* Function Name:   realloc2_func()                                  */
/*********************************************************************/
int realloc2_func(void)
{
    int size;
    int num;
    void *malloc_ptr = NULL;
    void *calloc_ptr = NULL;
    void *realloc_ptr = NULL;

    PrintMsg("entering realloc2_func mixing 31 and 64 bit");

    // test case : malloc, realloc with 1K, free
    size = 150;
    if (!(malloc_ptr = getMalloc(size, TPF_HEAP_31)))
        return(1);
    size = SIZE_1K;
    if (!(realloc_ptr = getRealloc(malloc_ptr, size, TPF_HEAP_64)))
        return(1);
    FreeStorage(realloc_ptr);


    // test case : malloc, realloc with smaller size
    size = 1000;
    if (!(malloc_ptr = getMalloc(size, TPF_HEAP_64)))
        return(1);
    size = 100;
    if (!(realloc_ptr = getRealloc(malloc_ptr, size, TPF_HEAP_31)))
        return(1);
    FreeStorage(realloc_ptr);


    // test case : calloc, realloc 4K, free
    size = 1000;
    num  = 1;
    if (!(calloc_ptr = getCalloc(num, size, TPF_HEAP_31)))
        return(1);
    size = SIZE_4K;
    if (!(realloc_ptr = getRealloc(calloc_ptr, size, TPF_HEAP_64)))
        return(1);
    FreeStorage(realloc_ptr);


    // test case : calloc, realloc smaller, free
    size = 9000;
    num  = 1;
    if (!(calloc_ptr = getCalloc(num, size, TPF_HEAP_64)))
        return(1);
    size = 5000;
    if (!(realloc_ptr = getRealloc(calloc_ptr, size, TPF_HEAP_31)))
        return(1);
    FreeStorage(realloc_ptr);

    return(0);
}

/*********************************************************************/
/* Function Name:   free_func()                                      */
/*********************************************************************/
void free_func(void)
{
    PrintMsg("entering free_func");


    // Test Case:       Free ptr of NULL
    FreeStorage(NULL);

    // Test Case:       Free ptr of 0
    FreeStorage(0);


}


/**********************************************************************/

void *getMalloc(size_t size, enum t_heap heap)
{
    char *m_ptr = NULL;

    errno = 0;
    switch (heap) {
      case TPF_HEAP_31:
          m_ptr = (char *)malloc(size);
          break;

      case TPF_HEAP_64:
          m_ptr = (char *) malloc64(size);
          break;
    }

    if (m_ptr) {
        sprintf (msgBuffer, "Successful malloc: addr %lX %li bytes", (long)m_ptr, size);
        PrintMsg(msgBuffer);
    } else {
        if (!expectedError) {
            sprintf (msgBuffer, "Unsuccessful malloc: addr %lX %li bytes", (long)m_ptr, size);
            PrintError(msgBuffer);
        }
    }

    if (!expectedError)
        prterrno(errno);

    return (m_ptr);
}

void *getCalloc(size_t num, size_t size, enum t_heap heap)
{
    char *c_ptr = NULL;

    errno = 0;
    switch (heap) {
      case TPF_HEAP_31:
          c_ptr = (char *)calloc(num, size);
          break;

      case TPF_HEAP_64:
          c_ptr = (char *) calloc64(num, size);
          break;
    }

    if (c_ptr) {
        sprintf (msgBuffer, "Successful calloc: addr %lX %li elements %li bytes", (long)c_ptr, num, size);
        PrintMsg(msgBuffer);
    } else {
        if (!expectedError) {
            sprintf (msgBuffer, "Unsuccessful calloc: addr %lX %li elements %li bytes", (long)c_ptr, num, size);
            PrintError(msgBuffer);
        }
    }

    if (!expectedError)
        prterrno(errno);

    return (c_ptr);
}


int checkNulls (char *c_ptr, int num, int size)
{
    char *c2_ptr = NULL;
    char *last_byte = NULL;

    c2_ptr = strrchr (c_ptr, 0x00);
    if (c2_ptr != c_ptr) {
        PrintError("unsuccessful calloc: storage not initialized to zero");
        return (1);
    }

    last_byte = (char *)c_ptr + (num * size -1);
    memset(c_ptr, 0x7F, (num * size));

    if (*last_byte != 0x7F) {
        PrintError("unable to write or read from last byte in calloc area");
        return (1);
    }
    if (*c_ptr != 0x7F) {
        PrintError("unable to write or read from first byte in calloc area");
        return (1);
    }
    return (0);
}


void* getRealloc(void *realloc_ptr, size_t size, enum t_heap heap)
{
    void *r_ptr = NULL;

    errno = 0;
    switch (heap) {
      case TPF_HEAP_31:
          r_ptr = (char *)realloc(realloc_ptr, size);
          break;

      case TPF_HEAP_64:
          r_ptr = (char *) realloc64(realloc_ptr, size);
          break;
    }

    if (r_ptr) {
        sprintf (msgBuffer, "Successful realloc: addr %lX %li bytes", (long)r_ptr, size);
        PrintMsg(msgBuffer);
    } else {
        if (!expectedError) {
            sprintf (msgBuffer, "Unsuccessful realloc: addr %lX %li bytes, errno %d- %s",
                                 (long)r_ptr, size, errno, strerror(errno));
            PrintError(msgBuffer);
        }
    }

    if (!expectedError)
        prterrno(errno);

    return (r_ptr);
}


int writeToStorage (void *ptr, int size, int numbytes)
{
    char *last_byte;
    int value = 0xFF;

    last_byte = (char *)ptr + size -1;
    if (numbytes == 1) {
        *(last_byte) = value;
        if (*last_byte != value) {
            PrintError("unable to write to ecb storage");
            return(1);
        }
    } else {
        memset(ptr, value, size);
        if (*(last_byte) != value) {
            PrintError("unable to memset ecb storage");
            return(1);
        }
    }
    return (0);
}

int readFromStorage (void *ptr, int size)
{
    char *last_byte;

    last_byte = (char *)ptr + size - 1;
    if (*last_byte != 0xFF) {
        PrintError("unable to read from ecb storage");
        return(1);
    }
    return (0);
}

void FreeStorage(void *ptr)
{
    free (ptr);
    sprintf (msgBuffer, "Successful free: address %lX", (long)ptr);
    PrintMsg(msgBuffer);

}


int doErrorPaths(enum t_heap heap)
{
    void *storage_ptr = NULL;
    void *malloc_ptr = NULL;
    char *ptr = NULL;
    int size  = 0;
    int num   = 0;

    unsigned short peh   = 0;
    unsigned short emps  = 0;
    unsigned short xmmes = 0;

    struct ctka *keyptA = NULL;

    // go to BSS to get keypoint A values
    cebic_save();
    cebic_goto_bss();

    // retrieve keypoint A by tpf_getkc
    if (tpf_getkc('A', (void *)&keyptA, GETKC_NONE) != 0)
    {
       PrintError("Error retrieving keypoint A");
    }

    // Store the CTKA settings to local variables for use later
    emps   = keyptA->ck1emps;
    xmmes  = keyptA->ck1xmmes;
    peh    = keyptA->ck1mcpeh(keyptA->ck1mcinus);

    // we are done with keypoint A, free the malloc area now.
    free(keyptA);

    // get back to the DBI where we were.
    cebic_restore();

    switch (heap) {
      case TPF_HEAP_31:
          PrintMsg("entering error paths for 31 bit");
          break;
      case TPF_HEAP_64:
          PrintMsg("entering error paths for 64 bit");
          break;
    }


    // test case : malloc size of 0
    PrintMsg("attempting malloc with size 0");
    if ((storage_ptr = getMalloc(0, heap))) {
        PrintError("malloc did not fail with size of zero");
        FreeStorage(storage_ptr);
        return(1);
    }

    // test case : malloc size of -1
    PrintMsg("attempting malloc with size -1");
    if ((storage_ptr = (char *) getMalloc(-1, heap))) {
        PrintError("malloc did not fail with size of -1");
        FreeStorage(storage_ptr);
        return(1);
    }

    // test case : malloc size of INT_MAX
    PrintMsg("attempting malloc with INT_MAX");
    if ((storage_ptr = (char *) getMalloc(INT_MAX, heap))) {
        PrintError ("malloc did not fail with INT_MAX");
        FreeStorage(storage_ptr);
        return(1);
    }

    // malloc > max size
    switch (heap) {
      case TPF_HEAP_31:
           // maximum size of 31bit heap that can be obtained
           // is the pre-allocated ech heap size (CTKA -> PEH in 4KB)
           // plus the maximum size of ECB 31bit heap (CTKA -> EMPS in MB)
           size = emps * ONE_MEG;
           size += peh * FOUR_KB;
           break;
      case TPF_HEAP_64:
           size = xmmes * ONE_MEG;
           break;
    }

    ++size;
    PrintMsg ("attempting malloc with size greater than max ECB heap");
    if ((storage_ptr = getMalloc(size, heap))) {
        PrintError("malloc did not fail with greater than max ECB heap");
        FreeStorage(storage_ptr);
        return(1);
    }

    PrintMsg ("attempting calloc with size greater than max ECB heap");
    if ((storage_ptr = getCalloc(1, size, heap))) {
        PrintError("calloc did not fail with greater than max ECB heap");
        FreeStorage(storage_ptr);
        return(1);
    }

    PrintMsg ("attempting realloc with size greater than max ECB heap");
    malloc_ptr = getMalloc(10, heap);
    if ((storage_ptr = getRealloc(malloc_ptr, size, heap))) {
        PrintError("realloc did not fail with greater than max ECB heap");
        FreeStorage(storage_ptr);
        return(1);
    }
    FreeStorage(malloc_ptr);

    // test case : calloc size of INT_MAX
    PrintMsg ("attempting calloc with INT_MAX");
    if ((storage_ptr = (char *) getCalloc(1, INT_MAX, heap))) {
        PrintError("calloc did not fail with INT_MAX");
        FreeStorage(storage_ptr);
        return(1);
    }

    // test case : calloc size of 0, number 0
    PrintMsg ("attempting calloc with size/num of zero");
    size = 0;
    num = 0;
    if ((storage_ptr = (char *) getCalloc(size, num, heap))) {
        PrintError("calloc did not fail with 0/0");
        FreeStorage(storage_ptr);
        return(1);
    }

    // test case : calloc size of -1, number 1
    PrintMsg("attempting calloc with size of -1");
    size = -1;
    num = 1;
    if ((storage_ptr = (char *) getCalloc(size, num, heap))) {
        PrintError("calloc did not fail with size of -1");
        FreeStorage(storage_ptr);
        return(1);
    }

    // test case : calloc size of 1, number 0
    PrintMsg("attempting calloc with number elements 0");
    size = 15;
    num = 0;
    if ((storage_ptr = (char *) getCalloc(size, num, heap))) {
        PrintError("calloc did not fail with number of elements 0");
        FreeStorage(storage_ptr);
        return(1);
    }

    // test case : realloc size of -1
    PrintMsg ("attempting realloc with -1");
    if (!(malloc_ptr = malloc(1000)))
        return(1);
    if ((storage_ptr = (char *) getRealloc(malloc_ptr, -1, heap))) {
        PrintError ("realloc did not fail with size -1");
        FreeStorage(storage_ptr);
        return(1);
    }


    // test case : realloc size of INT_MAX
    PrintMsg ("attempting realloc with INT_MAX");
    if (!(malloc_ptr = malloc(1000)))
        return(1);
    if ((storage_ptr = (char *) getRealloc(malloc_ptr, INT_MAX, heap))) {
        PrintError ("realloc did not fail with INT_MAX");
        FreeStorage(storage_ptr);
        return(1);
    }


    // test case : realloc size 0, ptr not NULL
    PrintMsg ("attempting realloc with size 0, ptr not NULL");
    ptr = (char *)malloc(1000);
    if ((storage_ptr = getRealloc(ptr, 0, heap))) {
        PrintError("realloc did not fail with size 0, ptr not NULL");
        return(1);
    }


    // test cases that actually occur in QM5B

    // test case : realloc with invalid address   DUMPS
    CreateECB_int (REALLOC_ECB_HEAP, 1);

    // test case : realloc with freed address   DUMPS
    CreateECB_int (REALLOC_ECB_HEAP, 2);

    // test case : realloc and free original malloc ptr   DUMPS
    CreateECB_int (FREE_ECB_HEAP, 1);

    // test case : free invalid address of -1     DUMPS
    CreateECB_int (FREE_ECB_HEAP, 2);

    // test case : free embedded address          DUMPS
    CreateECB_int (FREE_ECB_HEAP, 3);

    // test case : malloc, write past ending address      DUMPS  (if in ecbheap check mode)
    CreateECB_int (MALLOC_ECB_HEAP, 1);

    // test case : malloc, write before starting address   DUMPS  (if in ecbheap check mode)
    CreateECB_int (MALLOC_ECB_HEAP, 2);

    return(0);
}

void CreateECB_int(t_task task, int genericNum)
{
    struct tpf_cresc_input tci;
    struct PassedParms parms, *p_ptr;
    char data[sizeof(struct PassedParms) * 2 + 1];
    int i, j;

    p_ptr = &parms;
    p_ptr->task = task;
    p_ptr->errorFlagSetting = errorFlag;
    p_ptr->verboseFlagSetting = verboseFlag;
    p_ptr->genericNum = genericNum;
    tci.program = memPgm;
    tci.istream = TPF_CRESC_IS_BALANCE;
    tci.wait = TPF_CRESC_WAIT_YES;
    tci.timeout = 10;

    // Turn PassedParms into a null terminated string
    // by expanding each byte to two bytes of non-null data
    for (i = 0, j = 0; i < sizeof(struct PassedParms); i++, j += 2) {
        data[j]   = ((char *)p_ptr)[i] | 0x0f;
        data[j+1] = ((char *)p_ptr)[i] | 0xf0;
    }
    data[j] = '\0'; // null terminate
    tci.data_length = sizeof(data);
    tci.data = &data;

    tpf_cresc(&tci);
}

void CreateECB_void(t_task task, void *genericPtr)
{
    struct tpf_cresc_input tci;
    struct PassedParms parms, *p_ptr;
    char data[sizeof(struct PassedParms) * 2 + 1];
    int i, j;

    p_ptr = &parms;
    p_ptr->task = task;
    p_ptr->errorFlagSetting = errorFlag;
    p_ptr->verboseFlagSetting = verboseFlag;
    p_ptr->genericPtr = genericPtr;
    tci.program = memPgm;
    tci.istream = TPF_CRESC_IS_BALANCE;
    tci.wait = TPF_CRESC_WAIT_YES;
    tci.timeout = 10;

    // Turn PassedParms into a null terminated string
    // by expanding each byte to two bytes of non-null data
    for (i = 0, j = 0; i < sizeof(struct PassedParms); i++, j += 2) {
        data[j]   = ((char *)p_ptr)[i] | 0x0f;
        data[j+1] = ((char *)p_ptr)[i] | 0xf0;
    }
    data[j] = '\0'; // null terminate
    tci.data_length = sizeof(data);
    tci.data = &data;

    tpf_cresc(&tci);

}

void CreateECB(t_task task)
{
    CreateECB_void(task, (void *)0);
}


void cgswbc_func()
{
    struct PassedParms *p_ptr;
    char ownername [32 + 1];
    char owner2 [33 + 1];

    // test case : cgswbc owner < 32 bytes
    memset(ownername, ' ', sizeof(ownername));
    strcpy(ownername, "owner less than 32");
    p_ptr = (struct PassedParms *)get_SWB(ownername);
    release_SWB(p_ptr);

    // test case : cgswbc owner > 32 bytes
    memset(owner2, ' ', sizeof(owner2));
    strcpy(owner2, "123456789012345678901234567890123 ");
    p_ptr = (struct PassedParms *)get_SWB(owner2);
    release_SWB(p_ptr);

    // test case : cgswbc owner = 32 bytes
    memset(ownername, ' ', sizeof(ownername));
    strcpy(ownername, "this owner string is equal to 32 ");
    p_ptr = (struct PassedParms *)get_SWB(ownername);
    release_SWB(p_ptr);

    // test case : cgswbc owner = blanks
    memset(ownername, ' ', sizeof(ownername));
    p_ptr = (struct PassedParms *)get_SWB(ownername);
    release_SWB(p_ptr);

    // test case : cgswbc owner = NULL
    memset(ownername, '0', 1);
    p_ptr = (struct PassedParms *)get_SWB(ownername);
    release_SWB(p_ptr);

    return;


}


void *get_SWB(char *owner)
{
    struct PassedParms *p_ptr;

    if ((p_ptr = (struct PassedParms *)CGSWBC(owner))) {
        sprintf (msgBuffer, "Successful CGSWBC: addr: %lX owner: %s", (long)p_ptr, owner);
        PrintMsg(msgBuffer);
    } else {
          sprintf (msgBuffer, "Unsuccessful CGSWBC: addr: %lX owner: %s", (long)p_ptr, owner);
          PrintError(msgBuffer);
          prterrno(errno);
    }

    return (p_ptr);
}


void release_SWB(void *swb_ptr)
{

    CRSWBC(swb_ptr);
    sprintf (msgBuffer, "Successful CRSWBC: addr: %lX", (long)swb_ptr);
    PrintMsg(msgBuffer);

    return;
}


void detac_levels(unsigned char *detach)
{
    if (levtest(D2)) {
        detac (D2);
        *detach^= LEVEL_2;
    }

    if (levtest(D3)) {
       detac (D3);
       *detach^= LEVEL_3;
    }

    if (levtest(D7)) {
        detac (D7);
        *detach^= LEVEL_7;
    }

    if (levtest(D9)) {
        detac (D9);
        *detach^= LEVEL_9;
    }

    if (levtest(DC)) {
        detac (DC);
        *detach^= LEVEL_C;
    }

    if (levtest(DF)) {
        detac (DF);
        *detach^= LEVEL_F;
    }
}

void attac_levels(unsigned char *detach)
{
    if ((*detach & LEVEL_2) && (ecbptr()->ce1dt2)) {
        crusa (1, D2);
        attac (D2);
    }

    if ((*detach & LEVEL_3) && (ecbptr()->ce1dt3)) {
        crusa (1, D3);
        attac (D3);
    }

    if ((*detach & LEVEL_7) && (ecbptr()->ce1dt7)) {
        crusa (1, D7);
        attac (D7);
    }

    if ((*detach & LEVEL_9) && (ecbptr()->ce1dt9)) {
        crusa (1, D9);
        attac (D9);
    }

    if ((*detach & LEVEL_C) && (ecbptr()->ce1dtc)) {
        crusa (1, DC);
        attac (DC);
    }

    if ((*detach & LEVEL_F) && (ecbptr()->ce1dtf)) {
        crusa (1, DF);
        attac (DF);
    }
}

void getcc_func()
{

    char *work1, *work2, *work3, *work4, *work5;
    struct am0sg *amsg1;

    //   test case:  getcc with record ID
    if ((amsg1 = (struct am0sg *) getcc(D2, GETCC_ATTR0, "OM"))) {
        sprintf (msgBuffer, "Successful getcc level 2, record ID - \"OM\" ");
        PrintMsg(msgBuffer);
    } else {
          sprintf (msgBuffer, "Unsuccessful getcc level 2, record ID - \"OM\" ");
          PrintError(msgBuffer);
          prterrno(errno);
    }

    //   test case:  getcc with L1 - 381 bytes
    if ((work1 = (char *) getcc(DC, GETCC_TYPE, L1))) {
        sprintf (msgBuffer, "Successful getcc level C, 381 bytes ");
        PrintMsg(msgBuffer);
    } else {
          sprintf (msgBuffer, "Unsuccessful getcc level C, 381 bytes ");
          PrintError(msgBuffer);
          prterrno(errno);
    }

    //   test case:  getcc common, 1100 bytes
    if ((work2 = (char *) getcc(DF, (enum t_getfmt) (GETCC_SIZE+GETCC_COMMON), 1100))) {
        sprintf (msgBuffer, "Successful getcc common level F, 1100 bytes ");
        PrintMsg(msgBuffer);
    } else {
          sprintf (msgBuffer, "Unsuccessful getcc common level F, 1100 bytes ");
          PrintError(msgBuffer);
          prterrno(errno);
    }

    //   test case:  getcc fill 1055 bytes blanks
    if ((work3 = (char *) getcc(D7, (enum t_getfmt) (GETCC_TYPE+GETCC_FILL), L2, ' '))) {
        sprintf (msgBuffer, "Successful getcc, blank fill, 1055 bytes ");
        PrintMsg(msgBuffer);
    } else {
          sprintf (msgBuffer, "Unsuccessful getcc, blank fill, 1055 bytes ");
          PrintError(msgBuffer);
          prterrno(errno);
    }

    //   test case:  getcc common and fill 4095 bytes blanks
    if ((work4 = (char *) getcc(D9, (enum t_getfmt) (GETCC_TYPE+GETCC_FILL+GETCC_COMMON), L4, ' '))) {
        sprintf (msgBuffer, "Successful getcc common, blank fill, 4095 bytes ");
        PrintMsg(msgBuffer);
    } else {
          sprintf (msgBuffer, "Unsuccessful getcc common, blank fill, 4095 bytes ");
          PrintError(msgBuffer);
          prterrno(errno);
    }

    //   test case:  getcc protected and fill 4095 bytes 0x00
    if ((work5 = (char *) getcc(D3, (enum t_getfmt) (GETCC_TYPE+GETCC_PROTECTED+GETCC_FILL), L4,0x00))) {
        sprintf (msgBuffer, "Successful getcc protected, zero fill, 4095 bytes ");
        PrintMsg(msgBuffer);
    } else {
          sprintf (msgBuffer, "Unsuccessful getcc protected, zero fill, 4095 bytes ");
          PrintError(msgBuffer);
          prterrno(errno);
    }

    //  relcc D7 and DC later, after doing ecbheap, systemheap, swbs, etc

    relcc(D2);
    relcc(D3);
    relcc(D9);
    relcc(DF);

    sprintf (msgBuffer, "Successful relcc: levels D2, D3, D9, DF ");
    PrintMsg(msgBuffer);
}

void relcc_func()
{

    if (levtest(D7))
        relcc(D7);

    if (levtest(DC))
        relcc(DC);

    sprintf (msgBuffer, "Successful relcc: levels D7, DC ");
    PrintMsg(msgBuffer);
}
